/*
mpfrcx.h

include file for polynomials over mpfr and mpc

Copyright (C) 2009 Andreas Enge

This file is part of the MPFRCX Library.

The MPFRCX Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or (at your
option) any later version.

The MPFRCX Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with the MPC Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA.
*/

#ifndef __MPFRCX_H
#define __MPFRCX_H
#include <stdio.h>
#include <stdlib.h>
#include "gmp.h"
#include "mpfr.h"
#include "mpc.h"

#define MPFRCX_VERSION_MAJOR      0
#define MPFRCX_VERSION_MINOR      2
#define MPFRCX_VERSION_PATCHLEVEL 0
#define MPFRCX_VERSION_STRING     "0.2.0"

typedef struct {
   int    size;
   /* the number of currently allocated coefficients. High level          */
   /* arithmetic functions may increase this number if more space is      */
   /* needed. Coefficients are only freed when the whole polynomial is.   */
   /* Allocated coefficients are immediately initialised with mpc_init.   */
   int   deg;
   /* A degree of -1 indicates the zero polynomial.                       */
   mp_prec_t prec;
   /* the precision used for initialising all the coefficients            */
   mpc_t *coeff;
}
__mpcx_struct;

typedef __mpcx_struct mpcx_t[1];
typedef __mpcx_struct *mpcx_ptr;
typedef __gmp_const __mpcx_struct *mpcx_srcptr;

typedef struct {
   int    size;
   int   deg;
   mp_prec_t prec;
   mpfr_t *coeff;
}
__mpfrx_struct;

typedef __mpfrx_struct mpfrx_t[1];
typedef __mpfrx_struct *mpfrx_ptr;
typedef __gmp_const __mpfrx_struct *mpfrx_srcptr;


/* Check if stdio.h is included */
#if defined (_GMP_H_HAVE_FILE)
#define _MPFRCX_H_HAVE_FILE 1
#endif

#if defined (__cplusplus)
extern "C" {
#endif
extern void mpcx_init (mpcx_ptr f, const int size, const mp_prec_t prec);
extern void mpcx_set (mpcx_ptr h, mpcx_srcptr f);
extern void mpcx_init_set (mpcx_ptr h, mpcx_srcptr f);
extern void mpcx_clear (mpcx_ptr f);
extern void mpcx_realloc (mpcx_ptr f, const int size);

extern void mpfrx_init (mpfrx_ptr f, const int size, const mp_prec_t prec);
extern void mpfrx_set (mpfrx_ptr h, mpfrx_srcptr f);
extern void mpfrx_init_set (mpfrx_ptr h, mpfrx_srcptr f);
extern void mpfrx_clear (mpfrx_ptr f);
extern void mpfrx_realloc (mpfrx_ptr f, const int size);

extern int mpcx_cmp (mpcx_srcptr f, mpcx_srcptr g);
extern void mpcx_urandom (mpcx_ptr f, int deg, gmp_randstate_t state);
extern int mpfrx_cmp (mpfrx_srcptr f, mpfrx_srcptr g);
extern void mpfrx_urandom (mpfrx_ptr f, int deg, gmp_randstate_t state);

extern void mpcx_add (mpcx_ptr h, mpcx_srcptr f, mpcx_srcptr g);
extern void mpcx_sub (mpcx_ptr h, mpcx_srcptr f, mpcx_srcptr g);
extern void mpcx_si_sub (mpcx_ptr h, const long int c, mpcx_srcptr f);
extern void mpcx_neg (mpcx_ptr h, mpcx_srcptr f);

extern void mpfrx_add (mpfrx_ptr h, mpfrx_srcptr f, mpfrx_srcptr g);
extern void mpfrx_sub (mpfrx_ptr h, mpfrx_srcptr f, mpfrx_srcptr g);
extern void mpfrx_si_sub (mpfrx_ptr h, const long int c, mpfrx_srcptr f);
extern void mpfrx_neg (mpfrx_ptr h, mpfrx_srcptr f);


extern void mpcx_mul (mpcx_ptr h, mpcx_srcptr f, mpcx_srcptr g);

extern void mpfrx_mul (mpfrx_ptr h, mpfrx_srcptr f, mpfrx_srcptr g);


extern void mpcx_rem (mpcx_ptr r, mpcx_srcptr f, mpcx_srcptr g);

extern void mpfrx_rem (mpfrx_ptr r, mpfrx_srcptr f, mpfrx_srcptr g);


extern void mpcx_reconstruct (mpcx_t rop, mpcx_t *factors, int no);
extern void mpcx_multieval (mpc_t *values, mpc_t *args, int no, mpcx_srcptr f);
extern void mpcx_root (mpc_t root, mpcx_t f);

extern void mpfrx_reconstruct (mpfrx_t rop, mpfrx_t *factors, int no);
extern void mpfrx_multieval (mpfr_t *values, mpfr_t *args, int no, mpfrx_srcptr f);
extern void mpfrx_root (mpfr_t root, mpfrx_t f);

#ifdef _MPFRCX_H_HAVE_FILE
extern size_t mpcx_out_str (FILE* stream, int base, size_t n_digits, mpcx_srcptr f);

extern size_t mpfrx_out_str (FILE* stream, int base, size_t n_digits, mpfrx_srcptr f);
#endif

#if defined (__cplusplus)
}
#endif

#define mpcx_get_deg(x) ((x)->deg)
#define mpfrx_get_deg(x) ((x)->deg)

#define mpcx_get_prec(x) (mpc_get_prec((x)->coeff[0]))
#define mpfrx_get_prec(x) (mpfr_get_prec((x)->coeff[0]))

#endif /* ifndef __MPFRCX_H */
